unit main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls;

type
  TfrmMain = class( TForm )
    btnInterpolate: TButton;
    imgColorbar: TImage;
    imgDataBicubic: TImage;
    procedure btnInterpolateClick( Sender: TObject );
  private
    { Private declarations }
  public
    { Public declarations }
    procedure DoBicubicInterpolation();
    procedure DrawDataBicubic();
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.dfm}

const
  DataCoarse: array[ 0..3, 0..3 ] of Double = ( ( 5, 4, 6, 1 ), ( 4, 2, 3, 2 ), ( 2, 1, 5, 4 ), ( 3, 5, 2, 1) );

var
  DataBicubic: array[ 0..512, 0..512 ] of Double;

function ConvertRange( OriginalRangeStart, OriginalRangeValue, OriginalRangeEnd, NewRangeStart, NewRangeEnd: Integer ): Double; overload;
var
  Ratio: Double;
begin
  Ratio := ( NewRangeEnd - NewRangeStart ) / ( OriginalRangeEnd - OriginalRangeStart );
  Result := ( OriginalRangeValue - OriginalRangeStart ) * Ratio + NewRangeStart;
end;

function ConvertRange( OriginalRangeStart, OriginalRangeValue, OriginalRangeEnd, NewRangeStart, NewRangeEnd: Double ): Double; overload;
var
  Ratio: Double;
begin
  Ratio := ( NewRangeEnd - NewRangeStart ) / ( OriginalRangeEnd - OriginalRangeStart );
  Result := ( OriginalRangeValue - OriginalRangeStart ) * Ratio + NewRangeStart;
end;

function Limit( Min, Value, Max: Integer ): Integer;
begin
  if Value > Max then
    Result := Max
  else if Value < Min then
    Result := Min
  else
    Result := Value;
end;

function Sample( u, v: Integer ): Double;
begin
  u := Limit( 0, u, 3 );
  v := Limit( 0, v, 3 );
  Result :=  DataCoarse[ u, v ];
end;

function CubicInterpolation( v0, v1, v2, v3, x: Double ): Double;
var
  P, Q, R, S: Double;
begin
  P := ( v3 - v2 ) - ( v0 - v1 );
  Q := ( v0 - v1 ) - P;
  R := v2 - v0;
  S := v1;
  Result := P * x * x * x + Q * x * x + R * x + S;
end;

procedure TfrmMain.btnInterpolateClick( Sender: TObject );
begin
  DoBicubicInterpolation();
  DrawDataBicubic();
end;

procedure TfrmMain.DoBicubicInterpolation();
var
  i, j: Integer;
  u, v: Double;
  int_u, int_v: Integer;
  frac_u, frac_v: Double;
  a00, a10, a20, a30: Double;
  a01, a11, a21, a31: Double;
  a02, a12, a22, a32: Double;
  a03, a13, a23, a33: Double;
  c0, c1, c2, c3: Double;
begin
  for j := 0 to 512 do
    begin
      v := ConvertRange( 0, j, 512, 0, 3 );
      int_v := Round( Int( v ) );
      frac_v := Frac( v );

      for i := 0 to 512 do
        begin
          u := ConvertRange( 0, i, 512, 0, 3 );
          int_u := Round( Int( u ) );
          frac_u := Frac( u );

          a00 := Sample( int_u - 1, int_v - 1 );
          a10 := Sample( int_u    , int_v - 1 );
          a20 := Sample( int_u + 1, int_v - 1 );
          a30 := Sample( int_u + 2, int_v - 1 );
          c0 := CubicInterpolation( a00, a10, a20, a30, frac_u );

          a01 := Sample( int_u - 1, int_v     );
          a11 := Sample( int_u    , int_v     );
          a21 := Sample( int_u + 1, int_v     );
          a31 := Sample( int_u + 2, int_v     );
          c1 := CubicInterpolation( a01, a11, a21, a31, frac_u );

          a02 := Sample( int_u - 1, int_v + 1 );
          a12 := Sample( int_u    , int_v + 1 );
          a22 := Sample( int_u + 1, int_v + 1 );
          a32 := Sample( int_u + 2, int_v + 1 );
          c2 := CubicInterpolation( a02, a12, a22, a32, frac_u );

          a03 := Sample( int_u - 1, int_v + 2 );
          a13 := Sample( int_u    , int_v + 2 );
          a23 := Sample( int_u + 1, int_v + 2 );
          a33 := Sample( int_u + 2, int_v + 2 );
          c3 := CubicInterpolation( a03, a13, a23, a33, frac_u );

          DataBicubic[ i, j ] := CubicInterpolation( c0, c1, c2, c3, frac_v );
        end;
    end;
end;

procedure TfrmMain.DrawDataBicubic();
var
  i, j: Integer;
  LowestValue, HighestValue: Double;
begin
  LowestValue := 1.7E308;
  HighestValue := -1.7E308;

  for j := 0 to 512 do
    for i := 0 to 512 do
      begin
        if ( DataBicubic[ i, j ] < LowestValue ) then
          LowestValue := DataBicubic[ i, j ];
        if ( DataBicubic[ i, j ] > HighestValue ) then
          HighestValue := DataBicubic[ i, j ];
      end;

  for j := 0 to 512 do
    for i := 0 to 512 do
      imgDataBicubic.Canvas.Pixels[ i, j ] := imgColorbar.Canvas.Pixels[ 0, Round( ConvertRange( LowestValue, DataBicubic[ i, j ], HighestValue, 512, 0 ) ) ];
end;

end.
